--- Guilder: Quick Share Module
-- This module allows the user to quickly share notes with other members in the group. If the note is a macro, it can easily be stored as well.
-- @class file
-- @name Modules/Share.lua
-- @release 1.1.0

if select(6, GetAddOnInfo("Guilder_" .. (debugstack():match("[o%.][d%.][u%.]les\\(.-)\\") or ""))) ~= "MISSING" then return end

local Guilder = _G.Guilder
if not Guilder then
  error("Guilder_Share requires Guilder")
end
local L = Guilder.L

local DEBUG = Guilder.DEBUG
local expect = Guilder.expect

-- default settings for the addon -------------------------------------------

local DATABASE_DEFAULTS = {
	player = "",
	title = "",
	note = "",
	location = "",
	macro = ""
}

-- declare the module -------------------------------------------------------

local Guilder_Share = Guilder:NewModule("Share", "AceEvent-3.0", "AceHook-3.0")
Guilder_Share:SetName(L["Quick Share"])
Guilder_Share:SetDescription(L["Shared notepad."])
Guilder_Share:SetDefaults(DATABASE_DEFAULTS)
if DEBUG then
	Guilder_Share:SetIcon([[Interface\AddOns\Guilder\Modules\Share\Share]])
else
	Guilder_Share:SetIcon([[Interface\AddOns\Guilder_Share\Share]])
end

-- variables ----------------------------------------------------------------

-- Standard Module Methods --------------------------------------------------

-- Is called when the module is loaded
-- @usage internal use only
function Guilder_Share:OnModuleLoaded(module)
	if not self.db.profile.enabled then return end
	local id = module.id
end

-- Is called when the module is enabled
-- @usage internal use only
function Guilder_Share:OnEnable()
	-- register group/raid changes
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
end

-- Is called when the module is disabled
-- @usage internal use only
function Guilder_Share:OnDisable()
	-- Called when the addon is disabled
	self:UnregisterAllEvents()
end

-- Is called when the raid roster changes
-- @usage internal use only
function Guilder_Share:RAID_ROSTER_UPDATE()
	if Guilder.configLoaded then
		self.options.div.args.who.values = Guilder.PlayersInGroup
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
	end
end

-- code reuse
Guilder_Share.PARTY_MEMBERS_CHANGED = Guilder_Share.RAID_ROSTER_UPDATE

-- Is called when the user opens the configuration screen
-- @usage internal use only
function Guilder_Share:UpdateConfigOptions()
	self.options.div.args.who.values = Guilder:PlayersInGroupOrGuild()
end

-- Is called when the module receives a message
-- @usage internal use only
function Guilder_Share:CommHandler(distribution, sender, command, ...)
	-- if module is disabled, do nothing
	if not self.db.profile.enabled then
		if command == "UQSN" then
			Guilder:ShowMessage(L["%s tried to send you a note, but Quick Share was disabled. The note has been discarded."]:format(sender))
			Guilder:SendComm(self.id, "WHISPER", sender, "RQSN", false)
		end
		return
	end
	-- only handle comms if they are not from yourself
	if sender == myName then return end
	local t = {}
	local n = select('#', ...)
	-- Update: Quick Share Note
	if command == "UQSN" then
		if n < 2 then
			if Guilder.db.profile.showDebug then
				Guilder:ShowMessage("UQSN was expecting at least 2 parameters!")
				Guilder:SendComm(self.id, "WHISPER", sender, "RQSN", false)
			end
			return
		end
		self:ShowNote(sender, ...)
	end
	-- Response: Quick Share Note
	if command == "RQSN" then
		if select(1, ...) then
			Guilder:ShowMessage(L["%s received your note!"]:format(sender))
		else
			Guilder:ShowMessage(L["%s's Quick Share is disabled! Note was not sent."]:format(sender))
		end
	end
end

-- General functions --------------------------------------------------------

-- Sends the note to the player using the "WHISPER" addon channel
-- @usage internal use only
function Guilder_Share:SendNote()
	local g = nil
	-- if module is disabled, do nothing
	if not self.db.profile.enabled then
		Guilder:ShowMessage(L["%s is disabled"]:format("|cff00c0ff"..L["Quick Share"].."|r"))
		return
	end
	if self.db.profile.player == "" then
		Guilder:ShowMessage(L["You have to select a player to send the note to!"])
		return
	end
	if Guilder:IsInRaid() then g = LibRaid end
	if Guilder:IsInParty() then g = LibParty end
	-- check member online
	if not (LibGuild:IsMemberOnline(self.db.profile.player) or (group and group:HasMember(self.db.profile.player) and group:IsOnline(self.db.profile.player))) then
		Guilder:ShowMessage(L["%s is not online!"]:format(self.db.profile.player))
		return
	end
	if self.db.profile.player == myName then
		Guilder:ShowMessage(L["You can't send a note to yourself!"])
		return
	end
	if not Guilder:GetUserVersion(self.db.profile.player) then
		Guilder:ShowMessage(L["It looks like %s doesn't have Guilder installed!"]:format(self.db.profile.player))
		return
	end
	if self.db.profile.player and self.db.profile.note ~= "" then
		Guilder:ShowMessage(L["Sending note to %s."]:format(self.db.profile.player))
		Guilder:SendComm(self.id, "WHISPER", self.db.profile.player, "UQSN",
			self.db.profile.title, self.db.profile.note, self.db.profile.location, self.db.profile.macro)
	end
end

-- Takes the received note and places the information in the fields, ready for the user to review.
-- Notifies the user that a note has been received. 
-- @param sender person who sent the note
-- @param title title of the note
-- @param note contents of the note
-- @param location whether it should be a Global or Character Specific Macro
-- @param macro suggested name for the macro
-- @usage internal use only
function Guilder_Share:ShowNote(sender, title, note, location, macro)
	self.db.profile.title = title
	self.db.profile.note = note
	self.db.profile.macro = macro
	if macro and macro ~= "" then
		self.db.profile.location = location
	end
	Guilder:ShowMessage(L["You have received a note from %s."]:format(sender).." "..
		L["It has been added to Quick Share for you to review."])
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
	Guilder:SendComm(self.id, "WHISPER", sender, "RQSN", true)
end

-- Saves a macro
-- @param name name of the macro
-- @param body contents of the macro
-- @usage internal use only
function Guilder_Share:MakeMacro(name, body)
	-- nil for general, 1 for character
	local index = GetMacroIndexByName(name)
	if index == 0 then
		index = CreateMacro(name, 1, body, self.db.profile.location, 1)
		Guilder:ShowMessage(L["Macro '%s' has been created."]:format(name))
	else
		EditMacro(index, name, 1, body)
		Guilder:ShowMessage(L["Macro '%s' has been updated."]:format(name))
	end
end

-- Checks if we can create the macro, then calls Guilder_Share:MakeMacro() to do the work
-- @usage internal use only
function Guilder_Share:CreateMacro()
	if self.db.profile.note == "" then
		Guilder:ShowMessage(L["Could not create the macro because the %s field is empty!"]:format(L["Note"]))
		return
	end
	if self.db.profile.macro == "" then
		Guilder:ShowMessage(L["Could not create the macro because the %s field is empty!"]:format(L["Macro Name"]))
		return
	end
	if Guilder:IsInCombat() then
		Guilder:ShowMessage(L["You are in combat! Macro will be created once you leave combat."])
		Guilder:OutOfCombatWrapper(self.MakeMacro, self.db.profile.macro, self.db.profile.note)
	else
		self:MakeMacro(self.db.profile.macro, self.db.profile.note)
	end
end
