--- Guilder: Option and Module handling
-- @class file
-- @name Options.lua
-- @release 2.0.x

local _G = _G
local Guilder = _G.Guilder
local L = Guilder.L

local DEBUG = Guilder.DEBUG
local expect = Guilder.expect

-- load additional libraries ------------------------------------------------

local AceConfig = LibStub and LibStub("AceConfig-3.0", true)
if not AceConfig then
	LoadAddOn("Ace3")
	AceConfig = LibStub and LibStub("AceConfig-3.0", true)
	if not AceConfig then
		message(("Guilder requires the library %q and will not work without it."):format("AceConfig-3.0"))
		error(("Guilder requires the library %q and will not work without it."):format("AceConfig-3.0"))
	end
end
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

-- variables ----------------------------------------------------------------

local global_functions = {}

Guilder.Options = {}
Guilder.options = {}

-- dictionary of module to defaults
local module_to_defaults = {}

local Module = {}
Guilder:SetDefaultModulePrototype(Module)

local recent_modules = {}

-- Hook events --------------------------------------------------------------

-- this event hook will make sure that the module option tables are loaded,
-- even if there is a delay in the loading process
LibStub("AceEvent-3.0"):RegisterEvent("ADDON_LOADED",
	function(event, addon)
		if not Guilder.Options or not Guilder.Options.HandleModuleLoad then
			return
		end
		while true do
			local module = table.remove(recent_modules, 1)
			if not module then
				break
			end
			-- add the options for the newly loaded module to the option panels
			Guilder.Options.HandleModuleLoad(module)
			-- tell all the other modules a new module was loaded
			Guilder:CallMethodOnModules("OnModuleLoaded", module)
		end
	end
)

-- Define option tables -----------------------------------------------------

-- Define the general configuration screen for Guilder
function Guilder.get_general_options()
	return
		"version", {
			name = L["Version: %s"]:format(Guilder.version),
			type = "description",
			order = -1,
			width = "normal"
		},
		"defaultArgs", {
			type = "group",
			name = L["Guilder"],
			args = {
				general = {
					type = "group",
					name = L["General Settings"],
					inline = true,
					order = 2,
					args = {
						desc = {
							type = "description",
							name = L["Use the options below to configure Guilder. If you are unsure what each option does, you can find out by hovering the mouse pointer over the option."],
							order = 0
						},
						minimap_icon = {
							type = "toggle",
							name = L["Minimap icon"],
							desc = L["Show the Guilder icon on the minimap."],
							order = 1,
							get = function(info)
								return not Guilder.db.profile.minimap_icon.hide
							end,
							set = function(info, value)
								Guilder.db.profile.minimap_icon.hide = not value
								if value then
									LibStub("LibDBIcon-1.0"):Show("Guilder")
								else
									LibStub("LibDBIcon-1.0"):Hide("Guilder")
								end
							end,
							hidden = function()
								return not LibStub("LibDataBroker-1.1", true) or not LibStub("LibDBIcon-1.0", true) or IsAddOnLoaded("Broker2FuBar")
							end
						},
						Hint = {
							type = "toggle",
							name = L["Show Hint"],
							desc = L["Show hint at the bottom of the tooltip"],
							order = 2,
							get = function() return Guilder.db.profile.info.showHint end,
							set = function(info, v) Guilder.db.profile.info.showHint = v end
						},
						Modules = {
							type = "toggle",
							name = L["Show Module State"],
							desc = L["Show module state in the tooltip"],
							order = 3,
							get = function() return Guilder.db.profile.info.showModules end,
							set = function(info, v) Guilder.db.profile.info.showModules = v end
						},
						Offline = {
							type = "toggle",
							name = L["Include Offline Officers"],
							desc = L["Include offline officers in the tooltip"],
							order = 4,
							get = function() return Guilder.db.profile.info.showOffline end,
							set = function(info, v) Guilder.db.profile.info.showOffline = v end
						},
						Online = {
							type = "toggle",
							name = L["Show Online Total"],
							desc = L["Show number of officers online"],
							order = 5,
							get = function() return Guilder.db.profile.info.showOnline end,
							set = function(info, v)
								Guilder.db.profile.info.showOnline = v
								Guilder:UpdateText()	
							end
						},
						Total = {
							type = "toggle",
							name = L["Show Total"],
							desc = L["Show total number of officers"],
							order = 6,
							get = function() return Guilder.db.profile.info.showTotal end,
							set = function(info, v)
								Guilder.db.profile.info.showTotal = v
								Guilder:UpdateText()
							end
						}
					}
				},
				ranks = {
					type = "group",
					name = L["Ranks"],
					inline = true,
					order = 3,
					args = {
						info = {
							type = "description",
							name = L["You are not currently in a guild!"]
						}
					}
				}
			}
		}
end

-- define the generic module options that each module shares
function Guilder.get_module_options()
	-- define the 'Modules' section
	local module_options = {
		type = 'group',
		name = L["Modules"],
		desc = L["Modules provide actual functionality for Guilder."],
		childGroups = "tree",
		args = {
			header = {
				type = "header",
				name = L["Modules"],
				desc = "",
				order = 1
			},
			description =  {
				type = "description",
				name = L["Modules provide actual functionality for Guilder. To view the different modules and change their settings, please select a module from the list on the left."],
				order = 2
			},
			div = {
				type = "header",
				name = "",
				desc = "",
				order = 3
			}
		}
	}
	-- define common module options
	local module_args = {
		enabled = {
			type = 'toggle',
			name = L["Enable"],
			desc = L["Globally enable this module."],
			get = function(info)
				return info.handler:IsEnabled()
			end,
			set = function(info, value)
				if value then
					Guilder:EnableModule(info.handler)
				else
					Guilder:DisableModule(info.handler)
				end
			end
		}
	}
	-- merge one table onto another
	local function merge_onto(dict, ...)
		for i = 1, select('#', ...), 2 do
			local k, v = select(i, ...)
			if not v.order then
				v.order = 100 + i
			end
			dict[k] = v
		end
	end
	-- creates the base options table for each module
	function Guilder.Options.modules_handle_module_load(module)
		local id = module.id
		
		local opt = {
			type = 'group',
			name = module.name,
			desc = module.description,
			icon = module.icon,
			args = {},
			handler = module,
		}
		module_options.args[id] = opt
		
		for k, v in pairs(module_args) do
			opt.args[k] = v
		end
		
		if global_functions[module] then
			merge_onto(opt.args, global_functions[module](module))
			global_functions[module] = false
		end
	end
	-- create the options table for each module
	for id, module in Guilder:IterateModules() do
		Guilder.Options.modules_handle_module_load(module)
	end
	-- and now for disabled modules not yet loaded
	local modules_not_loaded = Guilder.modules_not_loaded
	-- test if a module can be loaded
	local function loadable(info)
		local id = info[#info - 1]
		local _,_,_,_,loadable = GetAddOnInfo('Guilder_'..id)
		return loadable
	end
	-- test if a module can not be loaded
	local function unloadable(info)
		return not loadable(info)
	end
	-- create the 'Enable' checkbox
	local arg_enabled = {
		type = 'toggle',
		name = L["Enable"],
		desc = L["Globally enable this module."],
		get = function(info)
			return false
		end,
		set = function(info, value)
			local id = info[#info - 1]
			Guilder:LoadAndEnableModule(id)
		end,
		disabled = unloadable,
	}
	-- show a notice if a module isn't loaded
	local no_mem_notice = {
		type = 'description',
		name = L["This module is not loaded and will not take up and memory or processing power until enabled."],
		order = -1,
		hidden = unloadable,
	}
	-- show a notice is a module can not be loaded
	local unloadable_notice = {
		type = 'description',
		name = function(info)
			local id = info[#info - 1]
			local _,_,_,_,loadable,reason = GetAddOnInfo('Guilder_'..id)
			if not loadable then
				if reason then
					if reason == "DISABLED" then
						reason = L["Disabled in the Blizzard addon list."]
					else
						reason = _G["ADDON_"..reason]
					end
				end
				if not reason then
					reason = UNKNOWN
				end
				return format(L["This module can not be loaded: %s"],reason)
			end
		end,
		order = -1,
		hidden = loadable,
	}
	-- create a simple options table if the module isn't loaded
	for id in pairs(modules_not_loaded) do
		if not module_options.args[id] then
			local addon_name = 'Guilder_' .. id
			local title = GetAddOnMetadata(addon_name, "Title")
			local notes = GetAddOnMetadata(addon_name, "Notes")
			
			local name = title:match("%[(.*)%]")
			if not name then
				name = id
			else
				name = name:gsub("|r", ""):gsub("|c%x%x%x%x%x%x%x%x", "")
			end
			
			local opt = {
				type = 'group',
				name = name,
				desc = notes,
				args = {
					enabled = arg_enabled,
					no_mem_notice = no_mem_notice,
					unloadable_notice = unloadable_notice,
				},
			}
			
			module_options.args[id] = opt
		end
	end
	-- return the options table
	return module_options
end

-- Local functions ----------------------------------------------------------

local new, del = Guilder.new, Guilder.del

-- Set the db instance on the module with defaults handled
-- @param module module for which to add defaults
-- @param defaults table of defaults to add
-- @usage internal use only
--     fix_db_for_module(MyModule, ModuleDefaults)
local function fix_db_for_module(module, defaults)
	if not defaults then
		defaults = {}
	end
	local gldr_db = Guilder.db
	local db = gldr_db:RegisterNamespace(module.id, {
		profile = defaults
	})
	module.db = db

	-- AceDB-3.0 is supposed to keep the child profiles linked to the main db.
	-- However, if a profile change happens while the child namespace isn't
	-- registered then it won't be kept in sync with the main db's profile.
	-- A good example of this is seen with load on demand modules. Setup two
	-- profiles. Enable the module in one and then go to the other and disable
	-- it (the order matters). Once the module has been disabled reload the UI
	-- so you load back into the UI with the profile that doesn't have the
	-- module loaded. Now select the other profile. Without the following line
	-- the profile will change but the module will load with the namespace still
	-- set to the old profile. 
	gldr_db.SetProfile(db, gldr_db:GetCurrentProfile())

	if not db.profile.enabled then
		module:Disable()
	end
end

do
	-- we need to hook OnInitialize so that we can handle the database stuff for modules
	local old_Guilder_OnInitialize = Guilder.OnInitialize
	Guilder.OnInitialize = function(self)
		if old_Guilder_OnInitialize then
			old_Guilder_OnInitialize(self)
		end
		
		for module, defaults in pairs(module_to_defaults) do
			fix_db_for_module(module, defaults)
		end
		-- no longer need this
		module_to_defaults = nil
	end
end

-- Iterate through the list of modules and only return 'id, module' if the module is enabled
-- @param modules table of modules
-- @param id module id
-- @usage internal use only
--     return enabled_iter, self.modules, nil
local function enabled_iter(modules, id)
	local id, module = next(modules, id)
	if not id then
		-- we're out of modules
		return nil
	end
	if not module:IsEnabled() then
		-- skip disabled modules
		return enabled_iter(modules, id)
	end
	return id, module
end

-- Module prototype and General module options ------------------------------

--- Set the function to be called that will return a tuple of key-value pairs that will be merged onto the options table for the module options.
-- @name Module:SetGlobalOptionsFunction
-- @param func function to call
-- @usage MyModule:SetGlobalOptionsFunction(function(self)
--     return 'someOption', { name = "Some option", } -- etc
-- end)
function Guilder.defaultModulePrototype:SetGlobalOptionsFunction(func)
	if DEBUG then
		expect(func, 'typeof', 'function')
		expect(global_functions[self], '==', nil)
	end
	
	global_functions[self] = func
end

-- Module functions ---------------------------------------------------------

--- Set the module's database defaults.
-- This will cause module.db to be set.
-- @param defaults defaults on a per-profile basis. can be nil.
-- @usage MyModule:SetDefaults({ color = { 1, 0, 0, 1 } })
function Module:SetDefaults(defaults)
	if DEBUG then
		expect(defaults, 'typeof', 'table;nil')
	end
	
	if not Guilder.db then
		-- full addon not loaded yet
		module_to_defaults[self] = defaults
	else
		fix_db_for_module(self, defaults)
	end
end

--- Set the localised description of the module.
-- @param description the localised description of the module, as a full sentence, including a period at the end.
-- @usage MyModule:SetDescription("This does a lot of things.")
function Module:SetDescription(description)
	if DEBUG then
		expect(description, 'typeof', 'string')
	end
	
	self.description = description
end

--- Set the localised icon for the module.
-- @param icon the path to the icon to use
-- @usage MyModule:SetIcon([[Interface\Addons\MyModule\myicon]])
function Module:SetIcon(icon)
	if DEBUG then
		expect(icon, 'typeof', 'string')
	end
	
	self.icon = icon
end

--- Set the localised name of the module.
-- @param name the localised name of the module, with proper spacing, and in Title Case.
-- @usage MyModule:SetName("My Module")
function Module:SetName(name)
	if DEBUG then
		expect(name, 'typeof', 'string')
	end
	
	self.name = name
end

function Module:SetModuleOptions(options)
	if DEBUG then
		expect(options, 'typeof', 'string;table')
	end
	
end

-- Guilder Module functions -------------------------------------------------

-- Called when module is first created
-- @param module function reference to the module
-- @usage internal use only
function Guilder:OnModuleCreated(module)
	local id = module.moduleName
	if DEBUG then
		expect(id, 'typeof', 'string')
		expect(id, 'match', '^[A-Za-z_][A-Za-z0-9_]*$')
	end
	module.id = id
	self[id] = module
	
	recent_modules[#recent_modules+1] = module
end

--- Disable a module properly.
-- Unlike module:Disable(), this tracks it in the DB
-- @param module the module to disable
-- @usage Guilder:DisableModule(MyModule)
function Guilder:DisableModule(module)
	if not module:IsEnabled() then
		return
	end
	module.db.profile.enabled = false
	module:Disable()
	self:ShowMessage(L["%s Disabled"]:format(module.name))
end

--- Enable a module properly.
-- Unlike module:Enable(), this tracks it in the DB
-- @param module the module to enable
-- @usage Guilder:EnableModule(MyModule)
function Guilder:EnableModule(module)
	if module:IsEnabled() then
		return
	end
	module.db.profile.enabled = true
	module:Enable()
	self:ShowMessage(L["%s Enabled"]:format(module.name))
end

--- Iterate over all enabled modules
-- @usage for id, module in Guilder:IterateEnabledModules() do
--     doSomethingWith(module)
-- end
-- @return iterator which returns the id and module
function Guilder:IterateEnabledModules()
	return enabled_iter, self.modules, nil
end

-- Only loads the options tables if 'options' exists
-- @param module function reference to the module
-- @usage internal use only
function Guilder.Options.HandleModuleLoad(module)
	if not options then
		-- doesn't matter yet, it'll be caught in the real config opening.
		return
	end
	
	Guilder.Options.modules_handle_module_load(module)
end

-- configuration functions --------------------------------------------------

-- Create a seperate options table for the Blizzard interface with only a button
-- to call our customized interface (Ace3)
AceConfig:RegisterOptionsTable("Guilder_Bliz", {
	name = L["Guilder"],
	handler = Guilder,
	type = 'group',
	args = {
		config = {
			name = L["Standalone config"],
			desc = L["Open a standlone config window, allowing you to actually configure Guilder."],
			type = 'execute',
			func = function()
				Guilder.Options.OpenConfig()
			end
		}
	},
})
AceConfigDialog:AddToBlizOptions("Guilder_Bliz", "Guilder")

do
	-- register our slash commands, with and without capitals
	for i, cmd in ipairs { "/Guilder", "/Gldr" } do
		_G["SLASH_GUILDER" .. (i*2 - 1)] = cmd
		_G["SLASH_GUILDER" .. (i*2)] = cmd:lower()
	end
	-- register our slash command handler
	_G.hash_SlashCmdList["GUILDER"] = nil
	_G.SlashCmdList["GUILDER"] = function()
		return Guilder.Options.OpenConfig()
	end
end

-- On first call defines the configuration screen and redefines itself to only display it on subsequent calls
-- @usage internal use only
function Guilder.Options.OpenConfig()
	-- redefine it so that we just open up the pane next time
	function Guilder.Options.OpenConfig()
		-- update Guilder configuration options
		Guilder:UpdateConfigOptions()
		-- call modules' UpdateConfigOptions
		Guilder:CallMethodOnModules("UpdateConfigOptions")
		-- show the configuration screen
		AceConfigDialog:Open("Guilder")
	end
	-- guilder options table
	options = {
		name = L["Guilder"],
		handler = Guilder,
		type = 'group',
		args = {
		},
	}
	-- quick way to create a counter so that options have a unique order
	local new_order
	do
		local current = 0
		function new_order()
			current = current + 1
			return current
		end
	end

	-- get the general options
	local t = { Guilder.get_general_options() }
	Guilder.get_general_options = nil
	-- merge them with the guilder options table
	for i = 1, #t, 2 do
		local k, v = t[i], t[i+1]
		
		options.args[k] = v
		v.order = new_order()
	end

	-- if Development, add the Show Debug Information checkbox
	if DEBUG then
		options.args["xDebug"] = {
			type = "toggle",
			name = "Show Debug Information",
			desc = "",
			order = -1,
			get = function() return Guilder.db.profile.showDebug end,
			set = function(info, v) Guilder.db.profile.showDebug = v end
		}
	end

	-- add LibSink Output options
	options.args.defaultArgs.args.output = Guilder:GetSinkAce3OptionsDataTable()
	-- add an explanation, since this is not included in the library by default
	options.args.defaultArgs.args.output.args.a = {
		type = "description",
		name = L["You can use these options to specify where you want the warning output to be sent."],
		order = 0
	}
	-- make them inline, so they don't appear in the tree
	options.args.defaultArgs.args.output.inline = true

	-- add the 'Modules' section
	options.args.modules = Guilder.get_module_options()
	Guilder.get_module_options = nil
	options.args.modules.order = new_order()

	for id, module in Guilder:IterateModules() do
		options.args.modules.args["x"..id] = {
			type = "toggle",
			name = module.name,
			desc = L["Globally enable this module."],
			get = function()
				return module:IsEnabled()
			end,
			set = function(info, v)
				if v then
					Guilder:EnableModule(module)
				else
					Guilder:DisableModule(module)
				end
			end
		}
	end

	-- add the 'Profiles' section
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Guilder.db)
	options.args.profile.order = new_order()

	-- make options globally visible for dynamic updated options
	Guilder.options = options

	-- register the options with the AceConfig GUI
	AceConfig:RegisterOptionsTable("Guilder", Guilder.options)
	AceConfigDialog:SetDefaultSize("Guilder", 800, 570)

	-- make module options visible to each module
	for id, module in Guilder:IterateModules() do
		module.options = Guilder.options.args.modules.args[module.id].args
	end

	-- register an event for when a player leaves combat
	LibStub("AceEvent-3.0").RegisterEvent("Guilder.Options", "PLAYER_REGEN_ENABLED", function()
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
	end)
	-- register an event for when a player enters combat
	LibStub("AceEvent-3.0").RegisterEvent("Guilder.Options", "PLAYER_REGEN_DISABLED", function()
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
	end)

	-- tell everyone that the config options are loaded	
	Guilder.configLoaded = true

	-- update Guilder configuration options
	Guilder:UpdateConfigOptions()
	-- call modules' UpdateConfigOptions
	Guilder:CallMethodOnModules("UpdateConfigOptions")

	-- rewrite the funtion to the new shorter version
	return Guilder.Options.OpenConfig()
end
